﻿using System;
using System.Collections.Generic;
using MathNet.Numerics;
using MathNet.Numerics.LinearAlgebra;

namespace cmaes
{
	public class CmaParameters
	{
		public int N;					// number of objective variables/problem dimension.
		public double sigma = 0.5;		// coordinate-wise standard deviation (step size).
		public double stopfitness = 1e-10;  // stop if fitness < stopfitness (minimization)
	}

	public delegate double FitnessFunction(Vector v);

	#region - Comments from the original file (purecmaes.m). -
	//% (mu/mu_w, lambda)-CMA-ES 
	//% CMA-ES: Evolution Strategy with Covariance Matrix Adaptation for
	//% nonlinear function minimization. To be used under the terms of the
	//% GNU General Public License (http://www.gnu.org/copyleft/gpl.html).
	//% Copyright: Nikolaus Hansen, 2003-09. 
	//% e-mail: hansen[at]lri.fr
	//%
	//% This code is an excerpt from cmaes.m and implements the key parts
	//% of the algorithm. It is intendend to be used for READING and
	//% UNDERSTANDING the basic flow and all details of the CMA
	//% *algorithm*. Use the cmaes.m code to run serious simulations: it
	//% is longer, but offers restarts, far better termination options, 
	//% and supposedly quite useful output.
	//%
	//% URL: http://www.lri.fr/~hansen/purecmaes.m
	//% References: See end of file. Last change: December, 11, 2009

	//% ---------------------------------------------------------------  
	//%%% REFERENCES
	//%
	//% The equation numbers refer to 
	//% Hansen, N. and S. Kern (2004). Evaluating the CMA Evolution
	//% Strategy on Multimodal Test Functions.  Eighth International
	//% Conference on Parallel Problem Solving from Nature PPSN VIII,
	//% Proceedings, pp. 282-291, Berlin: Springer. 
	//% (http://www.bionik.tu-berlin.de/user/niko/ppsn2004hansenkern.pdf)
	//% 
	//% Further references:
	//% Hansen, N. and A. Ostermeier (2001). Completely Derandomized
	//% Self-Adaptation in Evolution Strategies. Evolutionary Computation,
	//% 9(2), pp. 159-195.
	//% (http://www.bionik.tu-berlin.de/user/niko/cmaartic.pdf).
	//%
	//% Hansen, N., S.D. Mueller and P. Koumoutsakos (2003). Reducing the
	//% Time Complexity of the Derandomized Evolution Strategy with
	//% Covariance Matrix Adaptation (CMA-ES). Evolutionary Computation,
	//% 11(1).  (http://mitpress.mit.edu/journals/pdf/evco_11_1_1_0.pdf).
	//% 
	#endregion

	/// <summary>
	/// 
	/// PureCMA-ES.NET class
	/// -----------------------------------------------------------------------
	/// 
	/// Class written using N. Hansen's pure CMA-ES Matlab variant
	/// (available at http://www.lri.fr/˜hansen/cmaes_inmatlab.html,
	/// see also references above from the original Matlab implementation).
	/// 
	/// The peculiarity of this implementation is approximate calculation
	/// of eigenvalues and eigenvectors via [Tools::EigenProblemApprx]
	/// method which has limited numer of iterations for QR decomposition
	/// ([eigenIterLimit]) and a limited precision (see [eps] variable in
	/// this method).
	/// 
	/// This is done to improve the speed of the algorithm, but can cause
	/// problems for ill-posed covariance matrix. I know that there are
	/// faster numerical methods to solve full eigenproblem, considering
	/// tri-diagonal matrices and Givens rotations, but I had very limited
	/// time and decided to go the simple way. Maybe later I'll do it right :).
	/// 
	/// For any comments and question considering this implementation, please
	/// write to:
	///		"Yury Tsoy" <yurytsoy@gmail.com>.
	/// 
	/// P.S. Yup. 'Standard' implementations of eigenvalues and eigenvectors
	/// calculation (from Math.Net and AlgLib (http://www.alglib.net/)) just
	/// didn't work and crashed frightenly often.
	/// 
	/// </summary>
	public class PureCMAES
	{
		public FitnessFunction feval;
		public List<double> MeanFitness { get; set; }
		public List<double> MinFitness { get; set; }

		/// <summary>
		/// weights = log(mu+1/2)-log(1:mu)'
		/// </summary>
		/// <param name="mu"></param>
		/// <returns></returns>
		private Vector GetWeights(int mu)
		{
			Vector res = new Vector(mu);

			double c1 = Math.Log(mu + 0.5);
			double sum = 0.0;
			for (int i = 0; i < mu; ++i)
			{
				//  weights = log(mu+1/2)-log(1:mu)'; % muXone array for weighted recombination
				res[i] = c1 - Math.Log(i + 1);
				sum += res[i];
			}

			// weights = weights/sum(weights);     % normalize recombination weights array
			sum = 1.0 / sum;
			for (int i = 0; i < mu; ++i) { res[i] *= sum; }

			return res;
		}

		private bool CheckMatrix (Matrix c)
		{
			if (c.Determinant() <= 0.0)
			{
				return false;
			}

			int size = c.RowCount;
			for (int i=0; i<size; ++i)
			{
				for (int j=i; j<size; ++j)
				{
					if (c[i, j] != c[j, i])
					{
						return false;
					}
				}
			}

			return true;
		}

		public void Run (CmaParameters parameters)
		{
			//Matrix test = Matrix.Diagonal(new Vector(10, 1.0));
			//Tools.EigenProblemApprx(test);

			#region - Init properties. -
			MeanFitness = new List<double>();
			MinFitness = new List<double>();
			#endregion

			int N = parameters.N;
			// returns Nx1 pseudorandom matrix with values uniformly distributed in [0;1]
			Vector xmean = Vector.Random(N); // objective variables initial point
			double sigma = parameters.sigma; // coordinate wise standard deviation (step size)
			double stopfitness = parameters.stopfitness; // stop if fitness < stopfitness (minimization)
			//double stopeval = 1000.0*N*N; // stop after [stopeval] number of function evaluations
			double stopeval = 10000.0; // stop after [stopeval] number of function evaluations
			int eigenIterLimit = 4000;

			#region - Strategy parameter setting: Selection. -

			double lambda = 4 + (int) (3.0*Math.Log(N)); // population size, offspring number
			int mu = (int) (lambda*0.5); // number of parents/points for recombination
			Vector weights = GetWeights(mu); // mu x 1 array for weighted recombination

			double temp = Tools.Sum(weights);
			double mueff = temp * temp / weights.SquaredNorm();
				//  mueff=sum(weights)^2/sum(weights.^2); % variance-effectiveness of sum w_i x_i
			#endregion

			#region - Strategy parameter setting: Adaptation. -
			double cc = (4 + mueff/N)/(N + 4 + 2*mueff/N); // time constant for cumulation for C
			double cs = (mueff + 2)/(N + mueff + 5); // t-const for cumulation for sigma control
			double c1 = 2/((N + 1.3)*(N + 1.3) + mueff); // learning rate for rank-one update of C
			double cmu = 2*(mueff - 2 + 1/mueff)/((N + 2)*(N + 2) + mueff); // and for rank-mu update
			double damps = 1 + 2*Math.Max(0, Math.Sqrt((mueff - 1)/(N + 1)) - 1) + cs; // damping for sigma  
			#endregion

			#region - Initialize dynamic (internal) strategy parameters and constants. -
			Vector pc = new Vector(N), ps = new Vector(N); // evolution paths for C and sigma
			Matrix B = Matrix.Identity(N, N); // B defines the coordinate system
			Vector D = Vector.Ones(N); // diagonal D defines the scaling

			Matrix Ddiag = Matrix.Diagonal(new Vector(Tools.Sqr(D).ToArray()));
			Matrix Btrans = new Matrix(B);
			Btrans.Transpose();
			Matrix C = B.Multiply(Ddiag).Multiply(Btrans);	// covariance matrix C
			Matrix Dinv = Matrix.Diagonal(Tools.Inv(D));
			Matrix invsqrtC = B.Multiply(Dinv).Multiply(Btrans); // C^-1/2 
			//Matrix invsqrtC = Tools.Mul(Tools.Mul(B, Dinv), Btrans); // C^-1/2 
			double eigeneval = 0.0; // track update of B and D
			double chiN = Math.Sqrt(N)*(1 - 1.0/(4*N) + 1.0/(21*N*N)); // expectation of ||N(0,I)|| == norm(randn(N,1))
			#endregion

			//--------------------------------------------------------------------

			#region - Generation Loop. -
			int counteval = 0;
			Matrix arx = new Matrix(N, (int)lambda);
			Vector arfitness = new Vector((int)lambda);
			int countex = 0;
			while (counteval < stopeval)
			{
				// Generate and evaluate lambda offsprings
				for (int k = 0; k < (int)lambda; ++k)
				{
					// randn -- works similar to rand but uses N(0;1) distribution
					//List<double> temp =
					Vector tempV = Tools.MulElementWise(D, Tools.RandN(N));
					tempV = Tools.Mul(B, tempV);
					arx.SetColumnVector(xmean + tempV*sigma, k);
							//arx(:,k) = xmean + sigma*B*(D.*randn(N, 1)); // m + sig * Normal(0,C) 

					arfitness[k] = feval(arx.GetColumnVector(k));	// objective function call
					++counteval;
				}

				// Sort by fitness and compute weighted mean into xmean
				List<int> arindex = new List<int>(new int[arfitness.Length]);
				for (int i = 0; i < arindex.Count; ++i) { arindex[i] = i; }
				Sorting.Sort(arfitness, arindex);	//[arfitness, arindex] = sort(arfitness);  % minimization
				Vector xold = xmean;

				#region - Save best and mean fitness values. -
				MeanFitness.Add(Tools.Mean(arfitness));
				MinFitness.Add(arfitness[0]); 
				#endregion

				// get [mu] best vectors.
				Matrix muBest = new Matrix(N, mu);
				for (int i=0; i<mu; ++i)
				{
					muBest.SetColumnVector(arx.GetColumnVector(arindex[i]), i);
				}
				// and perform their recombination.
				xmean = Tools.Mul(muBest, weights);
					//xmean = arx(:,arindex(1:mu)) * weights;  % recombination, new mean value

				//% Cumulation: Update evolution paths
				ps = (1-cs) * ps + Math.Sqrt(cs*(2-cs)*mueff) * Tools.Mul(invsqrtC, (xmean-xold)) / sigma;
				//double tempSig = ps.Norm()/(Math.Sqrt(1 - Math.Pow((1 - cs), (2*counteval/lambda))));
				int hsig = (ps.Norm()/(Math.Sqrt(1-Math.Pow((1-cs), (2*counteval/lambda)))*chiN) < 1.4 + 2.0/(N+1))
							?
							1 : 0;
				pc = (1-cc) * pc + hsig * Math.Sqrt(cc*(2-cc)*mueff) * (xmean-xold) / sigma; 

				//% Adapt covariance matrix C
				// repmat -- Replicate and tile [xold] array into [1 x mu] matrix
				Matrix artmp = (1/sigma) * (muBest - Tools.Repmat(xold,mu));
					//artmp = (1/sigma) * (arx(:,arindex(1:mu)) - repmat(xold,1,mu));
				C = (1-c1-cmu) * C				// regard old matrix  
				     + c1 * (Vector.DyadicProduct(pc, pc)			// plus rank one update
				     + (1-hsig) * cc*(2-cc) * C)					// correction if hsig==0
				     + cmu * artmp * Matrix.Diagonal(weights) * Matrix.Transpose(artmp);		// plus rank mu update 

				//% Adapt step size sigma
				sigma = sigma * Math.Exp((cs/damps)*(ps.Norm()/chiN - 1.0)); 

				//% Update B and D from C
				if (counteval - eigeneval > ((lambda/(c1+cmu))/N)*0.1)	//  to achieve O(N^2)
				{
					eigeneval = counteval;
					Tools.SimU(C);

					Tools.EigenProblemApprx(C, out D, out B, eigenIterLimit);

					D = Tools.Sqrt(D);			// D contains standard deviations now
					invsqrtC = B*Matrix.Diagonal( Tools.Inv(D)) * Matrix.Transpose(B);
				}

				//% Break, if fitness is good enough or condition exceeds 1e14, better termination methods are advisable 
				if (arfitness[0] <= stopfitness || Tools.Max(D) > 1e7 * Tools.Min(D))
					break;
			}
			MeanFitness.Add(Tools.Mean(arfitness));
			MinFitness.Add(arfitness[0]);
			#endregion
		}
	}
}
