﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using MathNet.Numerics;
using MathNet.Numerics.Distributions;
using MathNet.Numerics.LinearAlgebra;

namespace cmaes
{
	public class Tools
	{
		private static Random rand = new Random();
		private static NormalDistribution randn = new NormalDistribution(0, 1);

		#region - Basic stats and arithmetics for matrices and vectors. -
		public static List<double> Sqr(IList<double> mas)
		{
			List<double> res = new List<double>(new double[mas.Count]);
			for (int i = 0; i < mas.Count; ++i)
			{
				res[i] = mas[i] * mas[i];
			}
			return res;
		}

		public static List<List<double>> Sqr(List<List<double>> m)
		{
			List<List<double>> res = new List<List<double>>();
			int size = m.Count;
			for (int i = 0; i < size; ++i)
			{
				res.Add(new List<double>(new double [size]));
				for (int j = 0; j < size; ++j)
				{
					double temp = m[i][j];
					res[i][j] = temp*temp;
				}
			}
			return res;
		}

		public static Vector Sqrt (Vector v)
		{
			Vector res = v.Clone();
			for (int i = 0; i < res.Length; ++i)
			{
				res[i] = Math.Sqrt(res[i]);
			}
			return res;
		}

		public static List<List<double>> Inv (List<List<double>> m)
		{
			List<List<double>> res = new List<List<double>>();
			int size = m.Count;
			for (int i = 0; i < size; ++i)
			{
				res.Add(new List<double>(new double[size]));
				for (int j = 0; j < size; ++j)
				{
					res[i][j] = 1.0 / m[i][j];
				}
			}
			return res;
		}

		public static Vector Inv(Vector v)
		{
			Vector res = v.Clone();
			int size = v.Length;
			for (int i = 0; i < size; ++i)
			{
				res[i] = 1.0 / v[i];
			}
			return res;
		}

		public static double Sum(IList<double> mas)
		{
			double res = 0.0;
			foreach (var d in mas) { res += d; }
			return res;
		}

		public static double Mean(IList<double> mas)
		{
			return Sum(mas)/mas.Count;
		}

		public static int Max(IList<int> v)
		{
			int res = 0;
			foreach (int a in v)
			{
				if (res < a)
				{
					res = a;
				}
			}
			return res;
		}

		public static double Max(IList<double> v)
		{
			double res = 0.0;
			foreach (double a in v)
			{
				if (res < a)
				{
					res = a;
				}
			}
			return res;
		}

		public static double Min(IList<double> v)
		{
			double res = double.MaxValue;
			foreach (double a in v)
			{
				if (res > a)
				{
					res = a;
				}
			}
			return res;
		}
		#endregion

		#region - Generation of matrices. -
		public static List<List<double>> Rand(int height, int width) 
		{
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
				for (int j = 0; j < width; ++j) 
				{
					res[i][j] = rand.NextDouble();
				}
			}
			return res;
		}

		public static List<List<double>> RandN(int height, int width)
		{
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
				for (int j = 0; j < width; ++j)
				{
					res[i][j] = randn.NextDouble();
				}
			}
			return res;
		}

		public static Vector RandN(int length)
		{
			Vector res = new Vector(length);
			for (int i = 0; i < length; ++i)
			{
				res[i] = randn.NextDouble();
			}
			return res;
		}

		public static List<List<double>> Zeros(int height, int width)
		{
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
			}
			return res;
		}

		/// <summary>
		/// Returns identity matrix.
		/// </summary>
		/// <param name="height"></param>
		/// <param name="width"></param>
		/// <returns></returns>
		public static List<List<double>> Eye(int height, int width)
		{
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
				res[i][i] = 1.0;
			}
			return res;
		}

		/// <summary>
		/// Returns matrix filled with units.
		/// </summary>
		/// <param name="height"></param>
		/// <param name="width"></param>
		/// <returns></returns>
		public static List<List<double>> OnesEx(int height, int width)
		{
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
				for (int j = 0; j < width; ++j) { res[i][j] = 1.0; }
			}
			return res;
		}

		/// <summary>
		/// Returns matrix filled with units.
		/// </summary>
		/// <param name="height"></param>
		/// <param name="width"></param>
		/// <returns></returns>
		public static Matrix Ones(int height, int width)
		{
			Matrix res = new Matrix(height, width);
			for (int i = 0; i < height; ++i)
			{
				for (int j = 0; j < width; ++j) { res[i,j] = 1.0; }
			}
			return res;
		}

		public static List<List<double>> Diag(List<double> v)
		{
			int size = v.Count;
			List<List<double>> res = new List<List<double>>();
			for (int i = 0; i < size; ++i)
			{
				res.Add(new List<double>(new double[size]));
				res[i][i] = v[i];
			}
			return res;
		}

		public static List<List<double>> Diag(List<List<double>> m)
		{
			List<List<double>> res = new List<List<double>>();

			if (m.Count == 1)
			{	// [m] is a row vector.
				int size = m[0].Count;
				for (int i = 0; i < size; ++i)
				{
					res.Add(new List<double>(new double[size]));
					res[i][i] = m[0][i];
				}
			}
			else if (m[0].Count == 1)
			{	// [m] is a column vector.
				int size = m.Count;
				for (int i = 0; i < size; ++i)
				{
					res.Add(new List<double>(new double[size]));
					res[i][i] = m[i][0];
				}
			}
			else
			{	// m is a matrix.
				int size = m.Count;
				for (int i = 0; i < size; ++i)
				{
					res.Add(new List<double>(new double[size]));
					res[i][i] = m[i][i];
				}
			}

			return res;
		}

		public static Matrix Repmat (Vector v, int width)
		{
			Matrix res = new Matrix(v.Length, width);
			for (int i = 0; i < width; ++i )
			{
				res.SetColumnVector(v, i);
			}
			return res;
		}

		public static Matrix Triu (Matrix m)
		{
			Matrix res = m.Clone();
			int size = m.ColumnCount;
			for (int i = 1; i < size; ++i )
			{
				for (int j=0; j<i; ++j)
				{
					res[i,j] = 0;
				}
			}
			return res;
		}

		public static Matrix Triu1(Matrix m)
		{
			Matrix res = m.Clone();
			int size = m.ColumnCount;
			for (int i = 0; i < size; ++i)
			{
				for (int j = 0; j <= i; ++j)
				{
					res[i, j] = 0;
				}
			}
			return res;
		}

		public static void SimU(Matrix m)
		{
			int size = m.ColumnCount;
			for (int i = 1; i < size; ++i )
			{
				for (int j = 0; j < i; ++j )
				{
					m[i, j] = m[j, i];
				}
			}
		}
		#endregion

		#region - Search. -
		public static int IndexOf(Vector v, double val)
		{
			for (int i = 0; i < v.Length; ++i)
			{
				if (v[i] == val)
				{
					return i;
				}
			}
			return -1;
		} 
		#endregion

		#region - Arithmetic operations. -
		public static List<List<double>> Mul (List<List<double>> m1, List<List<double>> m2)
		{
			List<List<double>> res = new List<List<double>>();

			int height = m1.Count, width = m1[0].Count;
			for (int i = 0; i < height; ++i)
			{
				res.Add(new List<double>(new double[width]));
				for (int j = 0; j < width; ++j)
				{
					double temp = 0.0;
					for (int k = 0; k < height; ++k)
					{
						temp += m1[i][k] * m2[k][j];
					}
					res[i][j] = temp;
				}
			}

			return res;
		}

		public static List<List<double>> Mul(List<List<double>> m1, List<double> m2)
		{
			throw new NotImplementedException("cmaes :: Tools :: Mul");
		}

		public static List<List<double>> Transpose(List<List<double>> m)
		{
			List<List<double>> res = new List<List<double>>();

			int size = m.Count;
			for (int i = 0; i < size; ++i)
			{
				res.Add(new List<double>(new double[size]));
			}

			for (int i = 0; i < size; ++i)
			{
				res[i][i] = m[i][i];
				for (int j = i+1; j < size; ++j)
				{
					res[i][j] = m[j][i];
					res[j][i] = m[i][j];
				}
			}

			return res;
		}

		public static Vector MulElementWise(IVector<double> v1, IVector<double> v2)
		{
			Vector vres = new Vector(v1.Length);
			for (int i = 0; i < v1.Length; ++i )
			{
				vres[i] = v1[i] * v2[i];
			}
			return vres;
		}

		public static Vector Mul (Matrix m, Vector v)
		{
			Vector res = new Vector(m.RowCount);
			for (int i = 0; i < res.Length; ++i )
			{
				res[i] = Vector.ScalarProduct(m.GetRowVector(i), v);
			}
			return res;
		}

		#endregion

		#region - Eigenvalue decomposition and necessary methods. -
		public static void EigenProblemApprx(Matrix A, out Vector eigVal, out Matrix eigVec, int iterLimit)
		{
			int size = A.ColumnCount;

			Matrix T = A.Clone();
			double eps = 1e-4;
			//int iterCount = Math.Min(size * 10, 300);
			int iterCount = 0;
			do
			{
				QRDecomposition qrdecomp = T.QRDecomposition;
				T = qrdecomp.R * qrdecomp.Q;
				++iterCount;
			} while (!CheckTriU(T, eps) && iterCount < iterLimit);

			eigVal = new Vector(size);
			double eps_2 = eps * 0.5;
			Dictionary<double, int> lCount = new Dictionary<double, int>();
			for (int i = 0; i < size; ++i)
			{
				eigVal[i] = T[i, i];
				for (int j = 0; j < i; ++j )
				{
					if (Math.Abs(eigVal[i] - eigVal[j]) < eps_2) 
					{
						eigVal[i] = eigVal[j];
						break;
					}
				}

				if (lCount.ContainsKey(eigVal[i]))
				{
					++lCount[eigVal[i]];
				}
				else
				{
					lCount.Add(eigVal[i], 1);
				}
			}

			eigVec = Matrix.Identity(size, size);
			int count = 0;
			#region - Calculation of eigen-vectors. -
			//Matrix eigenVectors = new Matrix(size, size);
			foreach (var pair in lCount) 
			{
				Vector lambda = new Vector(size, pair.Key);
				//// Get matrix B of coefficients for the Bx = Ax - \lambda x = 0 system.
				Matrix diag = Matrix.Diagonal(lambda);
				Matrix B = A - diag;
				//int rang = B.Rank();
				//double det = B.Determinant(), detA = A.Determinant(), rang2 = det * detA * rang;
				Matrix X = GetSolutions(B, eps);
				Vector c1 = Tools.Mul(A, X.GetColumnVector(0)), c2 = X.GetColumnVector(0) * pair.Key;
				Vector c3 = c1 - c2;
				double dist = c3.Norm(), dis2 = dist;

				for (int i = 0; i < X.ColumnCount && count < size; ++i )
				{
					eigVec.SetColumnVector(X.GetColumnVector(i), count);
					++count;
				}
			}

			#endregion
		}

		public static Matrix GetSolutions(Matrix m, double eps)
		{
			// convert to triangle matrix using Jordan-Gauss transforms.
			Matrix A = m.Clone();
			int size = A.ColumnCount;
			for (int i=0; i<size-1; ++i) 
			{
				// todo: make pivoting!
				Vector row = A.GetRowVector(i);

				//
				// check if row[i] == 0
				if (row[i] == 0.0)
				{	// if row[i] == 0, then find the first row among the remaining ones with [i]-th element
					//	non equal to zero and swap rows.
					//	if no such rows exist, then continue.
					int nextRow = -1;
					for (int j = i+1; j < size; ++j )
					{
						if (A[j,i] != 0) 
						{
							nextRow = j;
							break;
						}
					}

					if (nextRow >= 0)
					{
						// swap i-th and nextRow rows.
						Vector tempRow = A.GetRowVector(nextRow).Clone ();
						A.SetRowVector(row, nextRow);
						A.SetRowVector(tempRow, i);
						row = A.GetRowVector(i);
					}
					else continue;
				}
				// normal flow always.
				for (int j = i + 1; j < size; ++j)
				{
					//
					// make A[j,i] = 0;
					//
					double coef = A[j, i] / A[i, i];
					Vector newRow = A.GetRowVector(j) + row.Scale(-coef);
					//RoundVector(newRow, eps);
					A.SetRowVector(newRow, j);

					if (Math.Abs(A[j, i]) < eps) { A[j, i] = 0; }
				}
			}
			for (int i = 0; i < A.RowCount; ++i )
			{
				Vector temp = A.GetRowVector (i);
				RoundVector(temp, 2.0*eps);
				A.SetRowVector(temp, i);
			}
			//int rank = A.Rank();
			int rank = GetRankTriU(A);	//, rankDif = rank2 - rank;

			Matrix res = new Matrix(size, size-rank);
			if (rank < size)
			{	// get fundamental system of solutions.
				//

				res = new Matrix(size, size - rank);
				for (int i = rank, idx = 0; i < size; ++i, ++idx)
				{
					res[i, idx] = 1.0;

					//
					// find other coefficients of the solution vector.
					for (int j=i-1; j>=0; --j) 
					{
						if (A[j, j] != 0) 
						{
							Vector rowJ = A.GetRowVector(j);
							double dotProd = rowJ.ScalarMultiply(res.GetColumnVector(idx));
							res[j, idx] = -dotProd / A[j, j];
						}
					}
				}
			}
			else 
			{
				// todo: sometimes this happens! Find out what to do!
				//res = new Matrix(size, 1);	// the solution is trivial.
				return new Matrix(size, 1);
				//throw new Exception("[GetSolutions]: Trivial solution found");
			}

			// normalize the solution vectors.
			for (int i = 0; i < res.ColumnCount; ++i)
			{
				double norm = res.GetColumnVector(i).Norm();
				Vector temp = res.GetColumnVector(i).Scale(1 / norm);
				res.SetColumnVector(temp, i);
			}

			return res;
		}

		public static void RoundVector (Vector v, double eps) 
		{
			int size = v.Length;
			for (int i = 0; i < size; ++i )
			{
				if (Math.Abs(v[i]) < eps) v[i] = 0.0;
			}
		}

		/// <summary>
		/// Calculate rank for upper triagonal matrix.
		/// </summary>
		/// <returns></returns>
		public static int GetRankTriU( Matrix m ) 
		{
			int size = m.RowCount;

			int count = 0;	// a number of linearly dependent rows.
			List<int> depRows = new List<int>();
			for (int i = 0; i < size - 1; ++i) 
			{
				if (depRows.Contains(i)) continue;

				Vector rowI = m.GetRowVector(i);
				for (int j = i + 1; j < size; ++j )
				{
					if (depRows.Contains(j)) continue;
					if (IsLinearlyDependent (rowI, m.GetRowVector (j))) 
					{
						depRows.Add (j);
						++count;
					}
				}
			}

			//
			// calculate a number of zero-filled rows.
			int zeroCount = 0;
			for (int i = 0; i < size; ++i)
			{
				if (IsZeroVector(m.GetRowVector (i))) ++zeroCount;
			}

			return size - count - zeroCount;
		}

		public static bool IsLinearlyDependent(Vector v1, Vector v2) 
		{
			double dotProd = v1.ScalarMultiply (v2);
			double norm1 = v1.Norm(), norm2 = v2.Norm();
			if (dotProd == 0 && (norm1 !=0 || norm2 != 0)) 
			{
				return false;
			}

			double cos = dotProd / (norm1 * norm2);
			if (cos == -1.0 || cos == 1.0) 
			{
				return true;
			}
			return false;
		}

		public static bool IsZeroVector(Vector v) 
		{
			int size = v.Length;
			for (int i = 0; i < size; ++i ) 
			{
				if (v[i] != 0.0) { return false; }
			}
			return true;
		}

		public static Vector EigenValues(Matrix A, double mexp) 
		{
			Matrix T = Tridiag(A);

			int size = A.ColumnCount;
			int q=0;
			while (q != size) 
			{
				// check mexp.
				for (int i=0; i<size-1; ++i) 
				{
					double s1 = mexp * (A[i,i] + A[i+1, i+1]);
					if (A[i + 1, i] <= s1) 
					{
						A[i + 1, i] = A[i, i + 1] = 0.0;
					}
				}

				// find p and q (see Golub, page 381, alg. 8.2.3)
			}

			Vector res = new Vector(size);
			return res;
		}

		public static bool CheckTriU(Matrix m, double eps) 
		{
			int size = m.ColumnCount;

			for (int i = 1; i < size; ++i )
			{
				for (int j=0; j<i; ++j) 
				{
					if (m[i,j] > eps) 
					{
						return false;
					}
				}
			}
			return true;
		}

		public static Matrix Tridiag(Matrix m) 
		{
			Matrix A = m.Clone();
			int n = m.ColumnCount;
			for (int k = 0; k < n - 2; ++k )
			{
				Vector h = GetSubvector(m.GetColumnVector(k), k+1, n-1);
				Vector v = HouseholderTransform(h);
				double vNorm = Vector.ScalarProduct(v, v);
				Vector vN = v.Scale (1.0/vNorm);

				Matrix subA = GetSubmatrix(A, k+1, n-1);
				Vector p = 2 * Tools.Mul (A, vN);
				Vector w = p - Vector.ScalarProduct(p, v) * vN;
				subA = subA - Vector.DyadicProduct(v, w) - Vector.DyadicProduct (w, v);

				SetSubmatrix(A, subA, k+1, k+1);
			}

			return A;
		}

		public static Vector GetSubvector(Vector x, int n, int m) 
		{
			int size = m - n + 1;
			Vector res = new Vector(size);
			for (int i = n, count=0; i <= m; ++i, ++count )
			{
				res[count] = x[i];
			}
			return res;
		}

		public static Matrix GetSubmatrix(Matrix x, int n, int m)
		{
			int size = m - n + 1;
			Matrix res = new Matrix(size, size);
			for (int i = n, count = 0; i <= m; ++i, ++count)
			{
				Vector tempV = GetSubvector (x.GetRowVector(i), n, m);
				res.SetRowVector(tempV, count);
			}
			return res;
		}

		public static void SetSubvector(Vector dest, Vector src, int n) 
		{
			int size = dest.Length;
			for (int i = n, count = 0; i < size; ++i, ++count )
			{
				dest[i] = src[count];
			}
		}

		public static void SetSubmatrix(Matrix dest, Matrix src, int n, int m) 
		{
			for (int i = n, count = 0; i <= m; ++i, ++count)
			{
				Vector dV = dest.GetRowVector(i);
				Vector sV = src.GetRowVector(i);
				SetSubvector(dV, sV, m);
			}
		}

		public static Vector HouseholderTransform(Vector x) 
		{
			Vector v = x.Clone();
			int n = v.Length;
			double mu = v.Norm();
			if (mu != 0.0) 
			{
				double beta = x[0] + mu * (x[0] >= 0? 1 : -1);
				v = v.Scale(1.0 / beta);
			}
			v[0] = 1.0;
			return v;
		}
		#endregion
	}
}
